import httpx
from json import JSONDecodeError


class RequestDataError(Exception):
    """For invalid http request data
    """
    def __init__(self, message: str, method: str) -> None:
        super().__init__(message)
        self.message = message
        self.method = method

    def __str__(self) -> str:
        return f"Passing Data error for {self.method.upper()} method. {self.message}"


class RequestParamsError(Exception):
    """For invalid http request params
    """
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        return f"Passing Parameters error for GET method. {self.message}"


# classic httpx.Response.raise_for_status() function, but with minor changes
# https://github.com/encode/httpx/blob/321d4aa5097fe7f24cdfed7191c44de589294780/httpx/_models.py#L1475
def raise_for_status(response: httpx.Response) -> None:
    """Raise the `HTTPStatusError` if one occurred.
    """
    request = response._request
    if request is None:
        raise RuntimeError(
            "Cannot call `raise_for_status` as the request "
            "instance has not been set on this response."
        )

    if response.is_success:
        return

    if response.has_redirect_location:
        message = (
            "{error_type} '{0.status_code} {0.reason_phrase}' for url '{0.url}'\n"
            "Redirect location: '{0.headers[location]}'\n"
            "For more information check: https://httpstatuses.com/{0.status_code}"
        )
    else:
        message = (
            "{error_type} '{0.status_code} {0.reason_phrase}' "
            "(see https://httpstatuses.com/{0.status_code}) "
            "for url '{0.url}' and '{0.request.method}' method.\n"
            "Request parameters:\n"
            "{parameters}\n"
            "Response:\n"
            "{response_json}\n"
        )

    status_class = response.status_code // 100
    error_types = {
        1: "Informational response",
        3: "Redirect response",
        4: "Client error",
        5: "Server error",
    }
    error_type = error_types.get(status_class, "Invalid status code")
    try:
        message = message.format(
            response,
            error_type=error_type,
            parameters=response.request.content.decode("utf-8"),
            response_json=response.json(),
        )
    except JSONDecodeError:
        raise httpx.DecodingError("The server returned non json data")
    raise httpx.HTTPStatusError(message, request=request, response=response)



import httpx
import re
import dataclasses
from pydantic import (
    BaseModel,
    validator,
)
from typing import (
    Dict,
    List,
    Union,
    Any,
    TYPE_CHECKING,
    TypeVar,
    Coroutine,
    Iterator,
)
from json import JSONDecodeError
from itertools import zip_longest
from .exceptions import RequestParamsError, RequestDataError, raise_for_status

if TYPE_CHECKING:
    from .api import Api

E = Union["EndpointId", "EndpointIdIterator"]
T = TypeVar("T")

KwargsType = Dict[str, Union[List[Dict[str, Any]], Dict[str, Any]]]
KwargsDict = Dict[str, Dict[str, Any]]


@dataclasses.dataclass
class EndpointBase:
    api: "Api"
    url: str
    endpoint: str

    async def _request(self, kwargs: Dict[str, Any]) -> httpx.Response:
        action = [*kwargs][0]
        endpoint = self.endpoint

        if action == "get":
            params = {
                "headers": {
                    "authorization": f"Token {self.api.token}",
                    "accept": "application/json;",
                },
                "params": kwargs[action],
            }
            err = RequestParamsError(
                f"{action.upper()} method must contain object id"
                ' in the {"id": 1} format',
            )
        else:
            params = {
                "headers": {
                    "authorization": f"Token {self.api.token}",
                    "Content-Type": "application/json;",
                },
            }
            err = RequestDataError(
                f"{action.upper()} method must contain object id"
                ' in the {"id": 1} format',
                action,
            )
            if action in ("patch", "put", "post"):
                params = {"json": kwargs[action], **params}
        if "{id}" in endpoint:
            try:
                id_ = f"{kwargs[action]['id']}"
                endpoint = re.sub(r"{id}", id_, endpoint)
            except (KeyError, TypeError):
                raise err

        req = await getattr(self.api.http_session, action)(
            f"{self.api.base_url}{endpoint}",
            **params,
        )
        return req

    async def request(self, kwargs: Dict[str, Any]) -> E:
        """Send http request

        Args:
            kwargs: dict with http request action + params/data.
                kwargs example: {"get": {"name": "some_name"}}

        Returns:
            EndpointIdIterator or EndpointId class object.

            EndpointIdIterator class object: If httpx.Response.json()
                contains many results. EndpointIdIterator is Iterator
                with EndpointId objects.
            EndpointId class object: If httpx.Response.json()
                contains a single result.
                EndpointId describes NetBox object (site, device, circuit, etc.)

        Raises:
            httpx._exceptions:
                See https://github.com/encode/httpx/blob/master/httpx/_exceptions.py
            RequestDataError: For invalid http request data
            RequestParamsError: For invalid http request params
            The difference between data and parameters - https://www.python-httpx.org/quickstart/
        """
        req = await self._request(kwargs)

        if req.status_code == 204 and "post" in kwargs:
            raise httpx.RequestError("Request allocation error")

        raise_for_status(req)

        return await EndpointIdIterator(
            api=self.api, url=self.url, endpoint=self.endpoint, response=req
        )()


class ValidateEndpoint(BaseModel):
    kwargs: KwargsType

    @validator("kwargs", pre=True, always=True)
    def set_default_kwargs(cls, v: KwargsType) -> KwargsType:
        return v or {"get": {}}

    @validator("kwargs")
    def check_requests(cls, v: KwargsType) -> KwargsType:
        for key in v:
            if key not in ("get", "put", "post", "patch", "delete"):
                raise ValueError(
                    "Available arguments: "
                    "'get', 'put', 'post', 'patch', 'delete'"
                )
        return v


@dataclasses.dataclass
class EndpointAsIterator(EndpointBase):
    kwargs: KwargsType = dataclasses.field(repr=False)

    def __post_init__(self) -> None:
        self.model = self.dict_generator(self.kwargs)

    def __iter__(self) -> "EndpointAsIterator":
        return self

    def __next__(self) -> Coroutine[Any, Any, E]:
        model = next(self.model)
        return self.request(model)

    @staticmethod
    def dict_generator(
        kwargs: KwargsType,
    ) -> Iterator[KwargsType]:
        for t in [*kwargs.items()]:
            d = dict([t])
            value = [*d.values()][0]
            if value:
                if isinstance(value, list):
                    for item in zip_longest(
                        d, *d.values(), fillvalue="".join(d)
                    ):
                        yield dict([item])
                else:
                    yield d
            else:
                yield {[*d][0]: {}}


@dataclasses.dataclass
class Endpoint(EndpointBase):
    """NetBox API endpoint object
    ('/dcim/devices', '/ipam/roles', ...)

    Args:
        api (anac.core.Api): Api class object
        url (str): NetBox url
        endpoint (str): NetBox API endpoint str ('/dcim/devices/', ...)

    What is Endpoint:
        In [1]: from anac import api
           ...:
           ...: a = api(
           ...:     "http://netbox/",
           ...:     token="api_token",
           ...: )
           ...: await a.openapi()

        In [2]: type(a.dcim_devices)
        Out[2]: anac.core.endpoint.Endpoint

        In [3]: type(a.circuits_circuits)
        Out[3]: anac.core.endpoint.Endpoint
    """

    async def __call__(
        self, **kwargs: Union[List[Dict[str, Any]], Dict[str, Any]]
    ) -> Union[E, "EndpointAsIterator"]:
        """Endpoint object is a coroutine
        for interacting with NetBox API endpoint.

                                                 EndpointIdAsIterator
                                                 /
                    EndpointIdIterator/EndpointId
                    /                            \
        Api-Endpoint                             EndpointId
                    \
                    EndpointAsIterator

        You can run each Endpoint object:
        - to send http request and get results
        - to get Iterator with EndpointBase.request coroutines for pending http requests
        and running in event loop with asyncio.gather or asyncio.as_completed

        Args:
            kwargs: list or dict with http request actions + params/data

        Returns:
            - EndpointId class object: Describes NetBox object (site, device, circuit,
                etc.)
            - EndpointIdIterator class object: Iterator with EndpointId objects
            - EndpointAsIterator class object: Iterator with EndpointBase.request
                coroutines for pending http requests and running in event loop
                with asyncio.gather or asyncio.as_completed

            This coroutine has simple logic and returns:
                - EndpointId or EndpointIdIterator if kwargs is a dict
                - EndpointAsIterator if kwargs is a list

        Raises:
            httpx._exceptions:
                See https://github.com/encode/httpx/blob/master/httpx/_exceptions.py
            RequestDataError: For invalid http request data
            RequestParamsError: For invalid http request params
            The difference between data and parameters -
            https://www.python-httpx.org/quickstart/

        Usage:
            - get 'test_device' NetBox '/dcim/devices/' object as EndpointId class
            object:
            In [1]: from anac import api
               ...:
               ...: a = api(
               ...:     "http://netbox/",
               ...:     token="token",
               ...: )
               ...: await a.openapi()

            In [2]: test_device = await a.dcim_devices(get={"name": "test"})

            In [3]: test_device
            Out[3]: EndpointId(api=Api, url='http://netbox/api',
            endpoint='/dcim/devices/')

            In [4]: test_device.name
            Out[4]: 'test'

            In [5]: test_device.id
            Out[5]: 4010

            In [6]: test_device.device_role
            Out[6]:
            {'id': 1,
             'url': 'http://netbox/api/dcim/device-roles/1/',
             'name': 'TEST',
             'slug': 'test'}

            In [7]: test_device.device_role.id
            Out[7]: 1

            - get NetBox '/dcim/devices/' objects with 'manufacturer id' = 1 as
            EndpointIdIterator class object:
                In [8]: test_devices = await a.dcim_devices(
                    ...:     get={"device_type": {"manufacturer": {"id": 1}},
                                  "limit": 2}
                    ...: )

                In [9]: len(test_devices)
                Out[9]: 2

                In [10]: for test_device in test_devices:
                    ...:     print(test_device.id)
                4010
                3988

                In [11]: test_devices[0]
                Out[11]: EndpointId(api=Api, url='http://netbox/api',
                endpoint='/dcim/devices/')

                In [12]: test_devices[1]
                Out[12]: EndpointId(api=Api, url='http://netbox/api',
                endpoint='/dcim/devices/')

                In [13]: test_devices[0].name
                Out[13]: 'test'

                In [14]: test_devices[1].name
                Out[14]: 'test1'

            - post 2 new NetBox '/dcim/devices/' objects with EndpointAsIterator class
            object:
                In [23]: pending_new_devices = await a.dcim_devices(
                    ...:     post=[
                    ...:         {"name": "test2", "device_role": 1, "site": 1,
                                  "device_type": 1, "status": 1},
                    ...:         {"name": "test3", "device_role": 1, "site": 1,
                                  "device_type": 1, "status": 1},
                    ...:     ]
                    ...: )

                In [24]: pending_new_devices
                Out[24]: EndpointAsIterator(api=Api, url='http://netbox/api',
                endpoint='/dcim/devices/')

                In [25]: import asyncio

                In [26]: new_devices = await asyncio.gather(*pending_new_devices)

                In [27]: new_devices
                Out[27]:
                [EndpointId(api=Api, url='http://netbox/api',
                endpoint='/dcim/devices/'),
                 EndpointId(api=Api, url='http://netbox/api',
                 endpoint='/dcim/devices/')]

                In [28]: for new_device in new_devices:
                    ...:     print(new_device.id)
                4074
                4075

                In [29]: new_devices[0]
                Out[29]: EndpointId(api=Api, url='http://netbox/api',
                endpoint='/dcim/devices/')

                In [30]: new_devices[0].name
                Out[30]: 'test2'

                In [31]: new_devices[1]
                Out[31]: EndpointId(api=Api, url='http://netbox/api',
                endpoint='/dcim/devices/')

                In [32]: new_devices[1].name
                Out[32]: 'test3'
        """
        kwargs = ValidateEndpoint(kwargs=kwargs).kwargs

        if len(kwargs) == 1 and isinstance([*kwargs.values()][0], dict):
            req = await self.request(kwargs)
        else:
            return EndpointAsIterator(
                api=self.api,
                url=self.url,
                endpoint=self.endpoint,
                kwargs=kwargs,
            )
        return req


class ValidateEndpointId(BaseModel):
    kwargs: KwargsDict

    @validator("kwargs", pre=True, always=True)
    def set_default_kwargs(cls, v: KwargsDict) -> KwargsDict:
        return v or {"get": {}}

    @validator("kwargs")
    def check_requests(cls, v: KwargsDict) -> KwargsDict:
        for key in v:
            if key not in ("get", "put", "patch", "delete"):
                raise ValueError(
                    "Available arguments: 'get', 'put', 'patch', 'delete'"
                )
        return v


class DictAttribute(dict):
    pass


@dataclasses.dataclass
class EndpointIdAsIterator:
    """Iterator with EndpointId objects

    EndpointIdAsIterator contains the results of modifying a single
    NetBox object (with endpoint id), using the EndpointId coroutine,
    running with multiple arguments. These results are iterated
    according to the order of the EndpointId arguments (see example below).

    Args:
        responses: EndpointId objects list

    What is EndpointIdAsIterator:
        In [1]: from anac import api
           ...:
           ...: a = api(
           ...:     "http://netbox/",
           ...:     token="api_token",
           ...: )
           ...: await a.openapi()

        In [2]: test_device = await a.dcim_devices(get={"name": "test"})

        In [3]: test_device
        Out[3]: EndpointId(api=Api, url='http://netbox/api', endpoint='/dcim/devices/')

        In [4]: test_device.name
        Out[4]: test

        In [5]: test_device = await test_device(
            ...:     put={
            ...:         "name": "supertest",
            ...:         "device_role": 2,
            ...:         "site": 2,
            ...:         "device_type": 2,
            ...:         "status": 2,
            ...:     },
            ...:     patch={"name": "megatest"},
            ...: )

        In [6]: test_device
        Out[6]: EndpointIdAsIterator()

        In [7]: list(test_device)
        Out[7]:
        [EndpointId(api=Api, url='http://netbox/api', endpoint='/dcim/devices/{id}/'),
         EndpointId(api=Api, url='http://netbox/api', endpoint='/dcim/devices/{id}/')]

        In [8]: test_device[0].name
        Out[8]: 'supertest'

        In [9]: test_device[1].name
        Out[9]: 'megatest'
    """
    responses: List[Any] = dataclasses.field(repr=False)

    def __post_init__(self) -> None:
        self._index = 0

    def __iter__(self) -> "EndpointIdAsIterator":
        return self

    def __getitem__(self, index: int) -> "EndpointId":
        return self.responses[index]

    def __next__(self) -> "EndpointId":
        if self._index >= len(self.responses):
            raise StopIteration
        response = self.responses[self._index]
        self._index += 1
        return response

    def __len__(self) -> int:
        return len(self.responses)


@dataclasses.dataclass
class EndpointId(EndpointBase):
    """NetBox API endpoint id object
    ('/dcim/devices/{id}', '/ipam/roles/{id}', ...).

    EndpointId object has all attributes of NetBox object.
    EndpointId object has .response attribute, containing httpx.Response object.

    Args:
        api (anac.core.Api): Api class object
        url (str): NetBox url
        endpoint (str): NetBox API endpoint str ('/dcim/devices/', ...)

    What is EndpointId:
        In [1]: from anac import api
           ...:
           ...: a = api(
           ...:     "http://netbox/",
           ...:     token="api_token",
           ...: )
           ...: await a.openapi()

        In [2]: test_device = await a.dcim_devices(get={"name": "test"})

        In [3]: type(test_device)
        Out[3]: anac.core.endpoint.EndpointId

        In [4]: test_device.id
        Out[4]: 4010
    """
    kwargs: Dict[str, Any] = dataclasses.field(repr=False)

    def __post_init__(self) -> None:
        self.generate_attrs(self, self.kwargs)

    def generate_attrs(self, obj: T, dict_: Dict[str, Any]) -> None:
        for key, value in dict_.items():
            if isinstance(value, dict):
                dict_obj = DictAttribute(value)
                setattr(obj, key.lower().replace(" ", "_"), dict_obj)
                self.generate_attrs(dict_obj, value)
            else:
                setattr(
                    obj,
                    key.lower().replace(" ", "_"),
                    value,
                )

    def __str__(self):
        return self.__class__.__name__

    def __getattr__(self, name: str) -> Any:
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            raise AttributeError(
                f"'{str(self)}' object has no attribute '{name}'"
            )

    async def __call__(
        self, **kwargs: Dict[str, Any]
    ) -> Union[E, "EndpointIdAsIterator"]:
        """EndpointId object is a coroutine
        for interacting with NetBox API endpoint id.

        You can run each EndpointId object:
        - to send http request and get results/new EndpointId objects
        - to send http requests and get Iterator with results/new EndpointId objects.
            All http requests in this coroutine will be run serially using 'await'.

                                                 EndpointIdAsIterator
                                                 /
                    EndpointIdIterator/EndpointId
                    /                            \
        Api-Endpoint                             EndpointId
                    \
                    EndpointAsIterator

        Args:
            kwargs: dict with http request actions + params/data

        Returns:
            - new EndpointId class object: Describes NetBox object (site, device,
                circuit, etc.)
            - EndpointIdAsIterator class object: Iterator with new EndpointId objects

        Raises:
            httpx._exceptions:
                See https://github.com/encode/httpx/blob/master/httpx/_exceptions.py
            RequestDataError: For invalid http request data
            RequestParamsError: For invalid http request params
            The difference between data and parameters -
            https://www.python-httpx.org/quickstart/

        Usage:
            In [1]: from anac import api
               ...:
               ...: a = api(
               ...:     "http://netbox/",
               ...:     token="api_token",
               ...: )
               ...: await a.openapi()

            In [2]: test_device = await a.dcim_devices(get={"name": "test"})

            In [3]: type(test_device)
            Out[3]: anac.core.endpoint.EndpointId

            In [4]: test_device.name
            Out[4]: test

            In [5]: test_device = await test_device(get={}, patch={"name": "testtest"})

            In [6]: test_device
            Out[6]: EndpointIdAsIterator()

            In [7]: list(test_device)
            Out[7]:
            [EndpointId(api=Api, url='http://netbox/api', endpoint='/dcim/devices/{id}/'),
             EndpointId(api=Api, url='http://netbox/api', endpoint='/dcim/devices/{id}/')]

            In [8]: test_device[0].name
            Out[8]: 'test'

            In [9]: test_device[1].name
            Out[9]: 'testtest'
        """
        self.endpoint = (
            f"{self.endpoint}"
            if "{id}" in self.endpoint
            else f"{self.endpoint}{'{id}/'}"
        )

        kwargs = ValidateEndpointId(kwargs=kwargs).kwargs

        new_kwargs: KwargsType = {
            key: {"id": self.id, **value} for key, value in kwargs.items()
        }

        if len(new_kwargs) == 1:
            return await self.request(new_kwargs)
        else:
            responses = []
            models = EndpointAsIterator.dict_generator(new_kwargs)
            for model in models:
                responses.append(await self.request(model))
            return EndpointIdAsIterator(
                responses=responses,
            )


@dataclasses.dataclass
class EndpointIdIterator:
    """Iterator with EndpointId objects

    EndpointIdIterator contains the multiple EndpointId objects,
    if Endpoint http request returns more than one result
    (httpx.Response.json()['results'])

    Args:
        api (anac.core.Api): Api class object
        url (str): NetBox url
        endpoint (str): NetBox API endpoint str ('/dcim/devices/', ...)
        response (httpx.Response): httpx.Response object

    What is EndpointIdIterator:
        In [1]: from anac import api
           ...:
           ...: a = api(
           ...:     "http://netbox/",
           ...:     token="api_token",
           ...: )
           ...: await a.openapi()

        In [2]: all_devices = await a.dcim_devices(get={"limit": 100})

        In [3]: all_devices
        Out[3]: EndpointIdIterator(api=Api, url='http://netbox/api',
        endpoint='/dcim/devices/')

        In [4]: len(all_devices)
        Out[4]: 100
    """
    api: "Api"
    url: str
    endpoint: str
    response: httpx.Response = dataclasses.field(repr=False)

    def __post_init__(self) -> None:
        self._index: int = 0
        self._responses: List[EndpointId] = []
        self.dict_data: Dict[str, Any] = {}
        self.list_data: List[Dict[str, Any]] = []

        httpx_models_response = {"response": self.response}
        try:
            self.list_data = self.response.json()["results"]
            if len(self.list_data) == 1:
                self.dict_data = {**httpx_models_response, **self.list_data[0]}
        except KeyError:
            self.dict_data = {**httpx_models_response, **self.response.json()}
        except JSONDecodeError:
            if self.response.request.method == "DELETE":
                self.dict_data = httpx_models_response
            else:
                raise httpx.DecodingError("The server returned non json data")

    async def __call__(self) -> E:
        """EndpointIdIterator object is a service coroutine
        for using with EndpointBase.request coroutine

                                                 EndpointIdAsIterator
                                                 /
                    EndpointIdIterator/EndpointId
                    /                            \
        Api-Endpoint                             EndpointId
                    \
                    EndpointAsIterator

        Returns:
            - EndpointId class object: Describes NetBox object (site, device, circuit,
                etc.)
            - EndpointIdIterator class object: Iterator with multiple EndpointId objects

        Raises:
            N/A
        """
        if self.dict_data:
            return EndpointId(
                api=self.api,
                url=self.url,
                endpoint=self.endpoint,
                kwargs=self.dict_data,
            )
        self._responses = [
            EndpointId(
                api=self.api, url=self.url, endpoint=self.endpoint, kwargs=data
            )
            for data in self.list_data
        ]
        return self

    def __iter__(self) -> "EndpointIdIterator":
        return self

    def __getitem__(self, index: int) -> "EndpointId":
        return self._responses[index]

    def __next__(self) -> "EndpointId":
        if self._index >= len(self._responses):
            raise StopIteration
        response = self._responses[self._index]
        self._index += 1
        return response

    def __len__(self) -> int:
        return len(self._responses)



from pkg_resources import get_distribution, DistributionNotFound
from anac.core.api import Api as api
from anac.core.exceptions import (
    RequestDataError,
    RequestParamsError,
    raise_for_status,
)

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass

__all__ = ("api", "RequestDataError", "RequestParamsError", "raise_for_status")



import httpx
import re
from pydantic.dataclasses import dataclass
from typing import Dict, Any, TypeVar
from .exceptions import raise_for_status
from .endpoint import Endpoint

A = TypeVar("A", bound="Api")


@dataclass
class Api:
    """The initial API object.

    Run openapi() coroutine/method to pull down openapi spec
    and dynamically creates attributes/endpoints.
    After that you can use 'tab' autocompletion
    in python interpreter

    Args:
        url (str): NetBox url
        token (str): NetBox API token

    Returns:
        Api object

    Raises:
        TypeError: If you forgot to specify 'url' or 'token' argument

    Usage:
        In [1]: from anac import api
           ...:
           ...: a = api(
           ...:     "https://your_netbox",
           ...:     token="your_api_token",
           ...: )
           ...: await a.openapi()
    """
    url: str
    token: str

    def __repr__(self) -> str:
        return self.__class__.__name__

    def __post_init_post_parse__(self) -> None:

        self.http_session = httpx.AsyncClient()

        self.base_url = f"{self.url if self.url[-1] != '/' else self.url[:-1]}/api"

    async def get_openapi(self, timeout: float) -> Dict[str, Any]:
        headers = {
            "Content-Type": "application/json;",
        }
        req = await self.http_session.get(
            f"{self.base_url}/docs/?format=openapi",
            headers=headers,
            timeout=timeout,
        )

        raise_for_status(req)

        return req.json()

    async def openapi(self, timeout: float = 50.0) -> None:
        """Get openapi spec and create attributes/endpoints
        with python interpreter autocompletion

        Args:
            timeout (float): Timeout for openapi http request

        Returns:
            N/A

        Raises:
            httpx.HTTPStatusError, httpx.ConnectError:
                See https://github.com/encode/httpx/blob/master/httpx/_exceptions.py
            json.JSONDecodeError: NetBox returns non json data

        Usage:
            In [1]: from anac import api
               ...:
               ...: a = api(
               ...:     "https://netbox",
               ...:     token="api_token",
               ...: )
               ...: await a.openapi()
        """
        self.open_api = await self.get_openapi(timeout=timeout)

        for endpoint in self.open_api["paths"].keys():
            setattr(
                self,
                re.sub(r"[-/]", "_", re.sub(r"[{}]", "", endpoint[1:-1])),
                Endpoint(self, self.base_url, endpoint),
            )

    async def aclose(self) -> None:
        """Close httpx.AsyncClient()

        Usage:
            In [1]: from anac import api
               ...:
               ...: a = api(
               ...:     "http://netbox",
               ...:     token="api_token",
               ...: )
               ...: await a.openapi()
               ...:

            In [2]: await a.aclose()
        """
        await self.http_session.aclose()

    async def __aenter__(self: A) -> A:
        """Use Api as a context manager

        Usage:
            In [1]: from anac import api
               ...:
               ...: async with api(
               ...:     "http://netbox",
               ...:     token="api_token",
               ...: ) as a:
               ...:     test1_device = await a.dcim_devices(get={"name": "test1"})

            In [2]: test1_device
            Out[2]: EndpointId(api=Api, url='http://your_netbox/api',
            endpoint='/dcim/devices/')
        """
        await self.openapi()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.http_session.aclose()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['anac', 'anac.core']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.22.0,<0.23.0', 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'anac',
    'version': '0.1.1',
    'description': 'Async NetBox API Client',
    'long_description': '[![Code style: black](https://img.shields.io/badge/code%20style-black-black?style=for-the-badge)](https://github.com/ambv/black)\n[![PyPI](https://img.shields.io/pypi/v/anac?style=for-the-badge)](https://pypi.org/project/anac)\n[![License: APACHE-2.0](https://img.shields.io/github/license/timeforplanb123/anac?style=for-the-badge)](https://opensource.org/licenses/Apache-2.0)\n[![Docs](https://img.shields.io/badge/docs-passing-green?style=for-the-badge)](https://timeforplanb123.github.io/anac/)\n\nanac\n==========\n\nPython **A**sync **N**etBox **A**PI **C**lient, based on <a href="https://github.com/encode/httpx" target="_blank">httpx</a> and <a href="https://github.com/samuelcolvin/pydantic" target="_blank">pydantic</a>\n\n\n## Documentation\n\n<a href="https://timeforplanb123.github.io/anac" target="_blank">https://timeforplanb123.github.io/anac</a>\n\n\n## Features \n\n* Minimalistic interface\n* Async only\n* Python interpreter autocompletion\n* Supports <a href="https://github.com/netbox-community/netbox" target="_blank">NetBox</a> 2.x, 3.x\n* Flexibility. All the objects are coroutines or coroutine iterators\n* Simple integration with parsers (<a href="https://github.com/google/textfsm" target="_blank">TextFSM</a>, <a href="https://github.com/dmulyalin/ttp" target="_blank">TTP</a>)\n\n\n## Quick Start \n\n#### Install\n\nPlease, at first, check the dependencies in `pyproject.toml` and create new virtual environment if necessary and then:\n\n**with pip:**\n\n```text\npip install anac \n```\n\n**with git:**\n\n```text\ngit clone https://github.com/timeforplanb123/anac.git\ncd anac \npip install .\n# or\npoetry install\n```\n\n\n#### Simple Examples\n\n\n#### Api Instantiating\n```python\nfrom anac import api\n\na = api(\n    "https://demo.netbox.dev",\n    token="cf1dc7b04de5f27cfc93aba9e3f537d2ad6fdf8c",\n)\n# get openapi spec and create attributes/endpoints   \nawait a.openapi()\n```\n#### `get` some device and `patch` it \n\n[![asciicast](https://asciinema.org/a/DmirYBxwl40VP9Delp6e0J3dE.svg)](https://asciinema.org/a/DmirYBxwl40VP9Delp6e0J3dE)\n\n```python\nIn [1]: some_device = await a.dcim_devices(get={"name": "dmi01-rochster-sw01"})\n\nIn [2]: some_device.name\nOut[2]: \'dmi01-rochster-sw01\'\n\nIn [3]: some_device.device_type\nOut[3]:\n{\'id\': 7,\n \'url\': \'https://demo.netbox.dev/api/dcim/device-types/7/\',\n \'display\': \'C9200-48P\',\n \'manufacturer\': {\'id\': 3,\n  \'url\': \'https://demo.netbox.dev/api/dcim/manufacturers/3/\',\n  \'display\': \'Cisco\',\n  \'name\': \'Cisco\',\n  \'slug\': \'cisco\'},\n \'model\': \'C9200-48P\',\n \'slug\': \'c9200-48p\'}\n\nIn [4]: some_device.status\nOut[4]: {\'value\': \'active\', \'label\': \'Active\'}\n\nIn [5]: some_device = await some_device(patch={"status": "failed"})\n\nIn [6]: some_device.status\nOut[6]: {\'value\': \'failed\', \'label\': \'Failed\'}\n```\n\n#### `get` some 2 devices and `put` + `patch` them\n```python\nIn [7]: some_devices = await a.dcim_devices(\n    ...:     get=[{"name": "dmi01-rochster-sw01"}, {"name": "dmi01-rochester-rtr01"}]\n    ...: )\n\n# EndpointAsIterator is a coroutine iterator with 2 coroutines\nIn [8]: some_devices\nOut[8]: EndpointAsIterator(api=Api, url=\'https://demo.netbox.dev/api\', endpoint=\'/dcim/devices/\')\n\nIn [9]: import asyncio\n\n# run 2 coroutines in the event loop\nIn [10]: some_devices = await asyncio.gather(*some_devices)\n\n# EndpointId is a NetBox \'/dcim/devices/\' object and coroutine\nIn [11]: some_devices\nOut[11]: \n[EndpointId(api=Api, url=\'https://demo.netbox.dev/api/dcim/devices/21/\', endpoint=\'/dcim/devices/\'),\n EndpointId(api=Api, url=\'https://demo.netbox.dev/api/dcim/devices/8/\', endpoint=\'/dcim/devices/\')]\n\nIn [12]: patch_some_devices = [coro(patch={"status": "failed"}) for coro in some_devices]\n\n# run 2 coroutines in the event loop\nIn [13]: patch_some_devices = await asyncio.gather(*patch_some_devices)\n\n# EndpointId is a coroutine, again\nIn [14]: patch_some_devices\nOut[14]: \n[EndpointId(api=Api, url=\'https://demo.netbox.dev/api/dcim/devices/21/\', endpoint=\'/dcim/devices/{id}/\'),\n EndpointId(api=Api, url=\'https://demo.netbox.dev/api/dcim/devices/8/\', endpoint=\'/dcim/devices/{id}/\')]\n\nIn [15]: patch_some_devices[0].name\nOut[15]: \'dmi01-rochster-sw01\'\n\nIn [16]: patch_some_devices[0].status\nOut[16]: {\'value\': \'failed\', \'label\': \'Failed\'}\n\nIn [17]: patch_some_devices[1].name\nOut[17]: \'dmi01-rochester-rtr01\'\n\nIn [18]: patch_some_devices[1].status\nOut[18]: {\'value\': \'failed\', \'label\': \'Failed\'}\n```\n\n#### `get` all devices\n```python\nIn [19]: all_devices = await a.dcim_devices(get={})\n# or\nIn [20]: all_devices = await a.dcim_devices()\n\n# EndpointIdIterator is an coroutine iterator with EndpointId objects\nIn [21]: all_devices\nOut[21]: EndpointIdIterator(api=Api, url=\'https://demo.netbox.dev/api\', endpoint=\'/dcim/devices/\')\n\nIn [22]: len(all_devices)\nOut[22]: 50\n\nIn [23]: all_devices[49]\nOut[23]: EndpointId(api=Api, url=\'https://demo.netbox.dev/api/dcim/devices/95/\', endpoint=\'/dcim/devices/\')\n\nIn [24]: all_devices[49].name\nOut[24]: \'ncsu118-distswitch1\'\n\n# by default, \'limit\' parameter = 50, but you can run \'get\' request with custom \'limit\'\nIn [25]: all_devices = await a.dcim_devices(get={"limit": 100})\n\nIn [26]: len(all_devices)\nOut[26]: 75\n\nIn [27]: all_devices[74]\nOut[27]: EndpointId(api=Api, url=\'https://demo.netbox.dev/api/dcim/devices/106/\', endpoint=\'/dcim/devices/\')\n\nIn [28]: all_devices[74].id\nOut[28]: 106\n```\n\n#### `get` all devices and `post` 2 new devices\n```python\nIn [29]: all_test = await a.dcim_devices(\n    ...:     get={},\n    ...:     post=[\n    ...:         {\n    ...:             "name": "test1",\n    ...:             "device_role": 1,\n    ...:             "site": 1,\n    ...:             "device_type": 1,\n    ...:             "status": "planned",\n    ...:         },\n    ...:         {\n    ...:             "name": "test2",\n    ...:             "device_role": 1,\n    ...:             "site": 1,\n    ...:             "device_type": 1,\n    ...:             "status": "planned",\n    ...:         },\n    ...:     ],\n    ...: )\n\n# run 3 coroutines in the event loop\nIn [30]: all_test = await asyncio.gather(*all_test)\n\n# EndpointIdIterator is an coroutine iterator with EndpointId objects\n# EndpointId is a NetBox \'/dcim/devices/\' object and coroutine\nIn [31]: all_test\nOut[31]: \n[EndpointIdIterator(api=Api, url=\'https://demo.netbox.dev/api\', endpoint=\'/dcim/devices/\'),\n EndpointId(api=Api, url=\'https://demo.netbox.dev/api/dcim/devices/110/\', endpoint=\'/dcim/devices/\'),\n EndpointId(api=Api, url=\'https://demo.netbox.dev/api/dcim/devices/111/\', endpoint=\'/dcim/devices/\')]\n\nIn [32]: all_test[0][49].name\nOut[32]: \'ncsu118-distswitch1\'\n\nIn [33]: all_test[1].name\nOut[34]: \'test1\'\n\nIn [35]: all_test[2].name\nOut[35]: \'test2\'\n\n# httpx.Response is available with .response attribute\nIn [36]: all_test[1].response.json()\nOut[36]:\n{\'id\': 110,\n \'url\': \'https://demo.netbox.dev/api/dcim/devices/110/\',\n \'display\': \'test1\',\n \'name\': \'test1\',\n \'device_type\': {\'id\': 1,\n  \'url\': \'https://demo.netbox.dev/api/dcim/device-types/1/\',\n  \'display\': \'MX480\',\n  \'manufacturer\': {\'id\': 7,\n  ...\n}\n```\n',
    'author': 'Pavel Shemetov',
    'author_email': 'pshemetov123@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/timeforplanb123/anac',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
